import numpy as np
import pytest
from bw2data import Database, Method, databases, geomapping, get_activity, methods
from bw2data.tests import bw2test

from simple_regional import OneSpatialScaleLCA as LCA


@bw2test
def test_value_error_no_method():
    empty = Database("empty")
    empty.write({("empty", "nothing"): {}})

    with pytest.raises(ValueError):
        LCA({("empty", "nothing"): 1})


@pytest.fixture
@bw2test
def import_data():
    biosphere_data = {
        ("biosphere", "F"): {
            "type": "emission",
            "exchanges": [],
        },
        ("biosphere", "G"): {
            "type": "emission",
            "exchanges": [],
        },
    }
    biosphere = Database("biosphere")
    biosphere.write(biosphere_data)

    inventory_data = {
        ("inventory", "U"): {
            "type": "process",
            "location": "L",
            "exchanges": [
                {"input": ("biosphere", "F"), "type": "biosphere", "amount": 1},
                {"input": ("biosphere", "G"), "type": "biosphere", "amount": 1},
            ],
        },
        ("inventory", "V"): {
            "type": "process",
            "location": "M",
            "exchanges": [],
        },
        ("inventory", "X"): {
            "type": "process",
            "location": "N",
            "exchanges": [],
        },
        ("inventory", "Y"): {
            "type": "process",
            "location": "O",
            "exchanges": [],
        },
        ("inventory", "Z"): {
            "type": "process",
            "location": "O",
            "exchanges": [],
        },
    }
    inventory = Database("inventory")
    inventory.write(inventory_data)

    method_data = [
        [("biosphere", "F"), 1, "L"],
        [("biosphere", "G"), 2, "L"],
    ]
    method = Method(("a", "method"))
    method.write(method_data)


def test_import_data(import_data):
    assert list(databases)
    assert list(methods)


@pytest.fixture
def lca(import_data):
    return LCA(demand={("inventory", "U"): 1}, method=("a", "method"))


def test_inventory(lca):
    lca.lci()

    assert np.allclose(lca.technosphere_matrix.todense(), np.eye(5))

    assert lca.biosphere_matrix.sum() == 2
    assert lca.biosphere_matrix.shape == (2, 5)
    assert (
        lca.biosphere_matrix[
            lca.dicts.biosphere[get_activity(("biosphere", "F")).id],
            lca.dicts.activity[get_activity(("inventory", "U")).id],
        ]
        == 1
    )
    assert (
        lca.biosphere_matrix[
            lca.dicts.biosphere[get_activity(("biosphere", "G")).id],
            lca.dicts.activity[get_activity(("inventory", "U")).id],
        ]
        == 1
    )

    # assert {("inventory", o) for o in "XUVYZ"} == set(lca.activity_dict.keys())
    # assert set(range(5)) == set(lca.activity_dict.values())
    # assert {("biosphere", "G"), ("biosphere", "F")} == set(
    #     lca.biosphere_dict.keys()
    # )
    # assert set(range(2)) == set(lca.biosphere_dict.values())

    assert lca.supply_array.sum() == 1
    assert lca.supply_array.shape == (5,)
    assert lca.supply_array[lca.dicts.product[get_activity(("inventory", "U")).id]] == 1


def test_characterization_matrix(lca):
    lca.lci()
    lca.lcia()

    print(list(lca.dicts.inv_spatial.items()))
    print(list(lca.dicts.biosphere.items()))
    print(list(geomapping.items()))
    for x in Database("biosphere"):
        print(x.id, x["code"])

    matrix = np.zeros((4, 2))
    matrix[1, 1] = 1
    matrix[1, 0] = 2
    print(lca.reg_cf_matrix.toarray())
    assert lca.reg_cf_matrix.sum() == 3
    assert lca.reg_cf_matrix.shape == (4, 2)
    assert (
        lca.reg_cf_matrix[
            lca.dicts.inv_spatial[geomapping["L"]],
            lca.dicts.biosphere[get_activity(("biosphere", "F")).id],
        ]
        == 1
    )
    assert (
        lca.reg_cf_matrix[
            lca.dicts.inv_spatial[geomapping["L"]],
            lca.dicts.biosphere[get_activity(("biosphere", "G")).id],
        ]
        == 2
    )


def test_lca_score(lca):
    lca.lci()
    lca.lcia()
    assert lca.score == 3



from functools import partial

import matrix_utils as mu
from bw2calc.lca import LCA
from bw2data import Database, Method, get_activity, methods
from bw_processing import load_datapackage, DatapackageBase
from fsspec.implementations.zip import ZipFileSystem
from fsspec import AbstractFileSystem
from typing import Callable, Iterable, Optional, Union


def dp(fp):
    return load_datapackage(ZipFileSystem(fp, mode="r"))


def get_dependent_databases(demand_dict):
    """Demand can be activitiy ids or tuple keys."""
    db_labels = [
        x[0] if isinstance(x, tuple) else get_activity(x)["database"]
        for x in demand_dict
    ]
    return set.union(*[Database(label).find_graph_dependents() for label in db_labels])


class OneSpatialScaleLCA(LCA):
    matrix_labels = [
        "biosphere_mm",
        "inv_mapping_mm",
        "reg_cf_mm",
        "technosphere_mm",
    ]

    def __init__(self, demand, *args, **kwargs):
        r"""Perform regionalized LCA calculation, where the inventory shares the same spatial scale as impact assessment.

        The calculation formula is:

        .. math::

            h_{r} = \left[ \textbf{MR} \right]^{T} \circ [ \textbf{B} \cdot (\textbf{A}^{-1}f) ]

        Uses sparse matrix `elementwise multiplication <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.multiply.html>`_.

        """
        self.databases = get_dependent_databases(demand)
        self.extra_data_objs = kwargs.pop("extra_data_objs", [])
        super(OneSpatialScaleLCA, self).__init__(demand, *args, **kwargs)
        if self.method not in methods:
            raise ValueError("Must pass valid `method` name")

    def create_inventory_mapping_matrix(self):
        """Get inventory mapping matrix, **M**, which maps inventory activities to inventory locations. Rows are inventory activities and columns are inventory spatial units.

        Uses ``self.technosphere_mm.row_mapper`` and ``self.databases``.

        Creates ``self.inv_mapping_mm``, ``self.inv_mapping_matrix``, and ``self.dicts.inv_spatial``/

        """
        self.inv_mapping_mm = mu.MappedMatrix(
            packages=[dp(Database(x).filepath_processed()) for x in self.databases]
            + self.extra_data_objs,
            matrix="inv_geomapping_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            row_mapper=self.technosphere_mm.col_mapper,
        )
        self.inv_mapping_matrix = self.inv_mapping_mm.matrix
        self.dicts.inv_spatial = partial(self.inv_mapping_mm.col_mapper.to_dict)

    def create_regionalized_characterization_matrix(self, row_mapper=None):
        """Get regionalized characterization matrix, **R**, which gives location- and biosphere flow-specific characterization factors.

        Rows are impact assessment spatial units, and columns are biosphere flows. However, we build it transverse and transpose it, as the characterization matrix indices are provided that way.

        Uses ``self._biosphere_dict`` and ``self.method``.

        Returns:
            * ``reg_cf_params``: Parameter array with row/col of IA locations/biosphere flows
            * ``ia_spatial_dict``: Dictionary linking impact assessment locations to matrix rows
            * ``reg_cf_matrix``: The matrix **R**

        """
        use_arrays, use_distributions = self.check_selective_use("characterization_matrix")

        try:
            self.reg_cf_mm = mu.MappedMatrix(
                packages=[dp(Method(self.method).filepath_processed())]
                + self.extra_data_objs,
                matrix="characterization_matrix",
                use_arrays=use_arrays,
                use_distributions=use_distributions,
                seed_override=self.seed_override,
                row_mapper=row_mapper,
                col_mapper=self.biosphere_mm.row_mapper,
                transpose=True,
            )
        except mu.errors.AllArraysEmpty:
            raise ValueError("Given `method` or `data_objs` have no characterization data")

        self.reg_cf_matrix = self.reg_cf_mm.matrix
        if row_mapper is None:
            self.dicts.ia_spatial = partial(self.reg_cf_mm.row_mapper.to_dict)

    def load_lcia_data(
        self,
        data_objs: Optional[Iterable[Union[AbstractFileSystem, DatapackageBase]]] = None
    ) -> None:
        self.create_inventory_mapping_matrix()
        self.create_regionalized_characterization_matrix(row_mapper=self.inv_mapping_mm.col_mapper)
        self.characterization_matrix = (
            self.inv_mapping_matrix @ self.reg_cf_matrix
        ).T

    def lcia_calculation(self):
        """Do regionalized LCA calculation.

        Creates ``self.characterized_inventory``.

        """
        # `.multiply()` does elementwise multiplication
        self.characterized_inventory = self.characterization_matrix.multiply(self.inventory)

    def results_ia_spatial_scale(self):
        raise NotImplementedError("No separate IA spatial scale")

    def results_inv_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return self.reg_cf_matrix.T.multiply(self.inventory * self.inv_mapping_matrix)



import numpy as np
from bw2data import Method, methods
from bw_processing import load_datapackage
from fsspec.implementations.zip import ZipFileSystem

# def import_regionalized_cfs(
#     geocollection,
#     method_tuple,
#     mapping,
#     scaling_factor=1,
#     global_cfs=None,
#     nan_value=None,
# ):
#     """Import data from a vector geospatial dataset into a ``Method``.

#     A ``Method`` can have both site-generic and regionalized characterization factors.

#     The ``mapping`` defines which field (vector) maps to which biosphere flows. Some methods may
#     only define regionalized chracterization factors for a single biosphere flow, but it is much
#     more common to have each field or band map to multiple biosphere flows. Therefore, mapping
#     should be defined as:

#     .. code-block:: python

#         {
#             field name (str): [list of biosphere flows (tuples)]
#         }

#     Args:
#         * *geocollection*: A ``geocollection`` name.
#         * *method_tuple*: A method tuple.
#         * *mapping*: Mapping from fields or bands to biosphere flows. See above.
#         * *scaling_factor*: Optional. Rescale the values in the spatial data source.
#         * *global_cfs*: An optional list of CFs to add when writing the method.
#         * *nan_value*: Sentinel value for missing values if ``NaN`` is not used directly.

#     """
#     assert (
#         geocollection in geocollections
#         and geocollections[geocollection].get("kind") == "vector"
#         and "field" in geocollections[geocollection]
#     )
#     gdf = gp.read_file(geocollections[geocollection]["filepath"])
#     id_label = geocollections[geocollection]["field"]

#     method = Method(method_tuple)
#     method.metadata["geocollections"] = [geocollection]
#     methods.flush()

#     data = []
#     if global_cfs:
#         data.extend(global_cfs)

#     for index, feature in gdf.iterrows():
#         for field_label, biosphere_flows in mapping.items():
#             value = feature[field_label]
#             if value is None or value == nan_value or np.isnan(value):
#                 continue
#             else:
#                 for flow in biosphere_flows:
#                     data.append(
#                         (
#                             flow,
#                             float(value) * scaling_factor,
#                             (geocollection, feature[id_label]),
#                         )
#                     )

#     method.write(data)


def dp(fp):
    return load_datapackage(ZipFileSystem(fp, mode="r"))
